@@ -114,11 +114,11 @@ static readstat_error_t readstat_begin_writing_data(readstat_writer_t *writer) {
                 goto cleanup;
         }
     }
+    writer->row_len = row_len;
+    writer->row = malloc(writer->row_len);
     if (writer->callbacks.begin_data) {
         retval = writer->callbacks.begin_data(writer);
     }
-    writer->row_len = row_len;
-    writer->row = malloc(writer->row_len);
 
 cleanup:
     return retval;
@@ -247,6 +247,21 @@ readstat_error_t readstat_write_spaces(readstat_writer_t *writer, size_t len) {
     return readstat_write_repeated_byte(writer, ' ', len);
 }
 
+readstat_error_t readstat_write_space_padded_string(readstat_writer_t *writer, const char *string, size_t max_len) {
+    readstat_error_t retval = READSTAT_OK;
+    if (string == NULL || string[0] == '\0')
+        return readstat_write_spaces(writer, max_len);
+
+    size_t len = strlen(string);
+    if (len > max_len)
+        len = max_len;
+
+    if ((retval = readstat_write_bytes(writer, string, len)) != READSTAT_OK)
+        return retval;
+
+    return readstat_write_spaces(writer, max_len - len);
+}
+
 readstat_label_set_t *readstat_add_label_set(readstat_writer_t *writer, readstat_type_t type, const char *name) {
     if (writer->label_sets_count == writer->label_sets_capacity) {
         writer->label_sets_capacity *= 2;
@@ -441,6 +456,11 @@ readstat_error_t readstat_writer_set_file_timestamp(readstat_writer_t *writer, t
     return READSTAT_OK;
 }
 
+readstat_error_t readstat_writer_set_table_name(readstat_writer_t *writer, const char *table_name) {
+    snprintf(writer->table_name, sizeof(writer->table_name), "%s", table_name);
+    return READSTAT_OK;
+}
+
 readstat_error_t readstat_writer_set_fweight_variable(readstat_writer_t *writer, const readstat_variable_t *variable) {
     if (readstat_variable_get_type_class(variable) == READSTAT_TYPE_CLASS_STRING)
         return READSTAT_ERROR_BAD_FREQUENCY_WEIGHT;
@@ -449,7 +469,7 @@ readstat_error_t readstat_writer_set_fweight_variable(readstat_writer_t *writer,
     return READSTAT_OK;
 }
 
-readstat_error_t readstat_writer_set_file_format_version(readstat_writer_t *writer, long version) {
+readstat_error_t readstat_writer_set_file_format_version(readstat_writer_t *writer, uint8_t version) {
     writer->version = version;
     return READSTAT_OK;
 }
