@@ -9,7 +9,7 @@ test_that("variable label stored as attributes", {
 test_that("value labels parsed from bcat file", {
   df <- read_sas("hadley.sas7bdat", "formats.sas7bcat")
 
-  expect_is(df$gender, "labelled")
+  expect_s3_class(df$gender, "haven_labelled")
   expect_equal(attr(df$gender, "labels"), c(Female = "f", Male = "m"))
   expect_equal(attr(df$workshop, "labels"), c(R = 1, SAS = 2))
 })
